/*
 * capnames.h : Definition of Capability Names
 *
 * This is an auto-generated file, do not edit
 */

#ifndef PROCPS_CAPNAMES_H
#define PROCPS_CAPNAMES_H

static const char *cap_names[] = {
    [0]  = "chown",
    [1]  = "dac_override",
    [2]  = "dac_read_search",
    [3]  = "fowner",
    [4]  = "fsetid",
    [5]  = "kill",
    [6]  = "setgid",
    [7]  = "setuid",
    [8]  = "setpcap",
    [9]  = "linux_immutable",
    [10]  = "net_bind_service",
    [11]  = "net_broadcast",
    [12]  = "net_admin",
    [13]  = "net_raw",
    [14]  = "ipc_lock",
    [15]  = "ipc_owner",
    [16]  = "sys_module",
    [17]  = "sys_rawio",
    [18]  = "sys_chroot",
    [19]  = "sys_ptrace",
    [20]  = "sys_pacct",
    [21]  = "sys_admin",
    [22]  = "sys_boot",
    [23]  = "sys_nice",
    [24]  = "sys_resource",
    [25]  = "sys_time",
    [26]  = "sys_tty_config",
    [27]  = "mknod",
    [28]  = "lease",
    [29]  = "audit_write",
    [30]  = "audit_control",
    [31]  = "setfcap",
    [32]  = "mac_override",
    [33]  = "mac_admin",
    [34]  = "syslog",
    [35]  = "wake_alarm",
    [36]  = "block_suspend",
    [37]  = "audit_read",
    [38]  = "perfmon",
    [39]  = "bpf",
    [40]  = "checkpoint_restore",
};


#define CAPABILITY_COUNT 40

#endif /* CAPNAMES_H */
